/*
anclock -- a simple analog clock
Copyright(c) 2009-2023 PT2K <webmaster@pt2k.xii.jp>

Permission to use, copy, modify, distribute, and sell this software and it's
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation.  No representations are made about the suitability of this
software for any purpose.  It is provided "as is" without express or
implied warranty.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include "screenhack.h"


#define MAX_OPT 1024

/* typedefs */
typedef unsigned short    ushort;
typedef unsigned long     ulong ;
typedef XWindowAttributes XGWA  ;
typedef struct {
  ushort r, g, b;
} RGB;

/* function macros */
#define i_random(i)  ((int)(random()%(i)))
#define i_round(x)   ((int)(((double)x)+0.5))
#define d_sin(x)     (sin(2*M_PI*(x)/360))
#define d_cos(x)     (cos(2*M_PI*(x)/360))
#define min(x,y)     (((x)<(y))?(x):(y))
#define average(x,y) (((x)+(y))/2)

/* commandline defaults */
static const char *anclock_defaults [] = {
  ".background:    black"  ,
  ".foreground:    white"  ,
  "*brightness:       70.0", /* [%]  */
  "*cycle:            20"  , /* [seconds] */
  "*face:          watch"  ,
  "*frequency:         0"  , /* [Hz] */
  "*full-minute:   False"  ,
  "*hide-index:    False"  ,
  "*hide-mark:     False"  ,
  "*hide-second:   False"  ,
  "*mono:          False"  ,
  "*pin:           False"  ,
  "*railway-clock: False"  ,
  "*size:             85.0", /* [%]  */
  "*wireframe:     False"  ,
  0
};

/* commandline options */
static XrmOptionDescRec anclock_options [] = {
  {"-background"   , ".background"   , XrmoptionSepArg,      0 },
  {"-brightness"   , ".brightness"   , XrmoptionSepArg,      0 },
  {"-cycle"        , ".cycle"        , XrmoptionSepArg,      0 },
  {"-face"         , ".face"         , XrmoptionSepArg, "watch"},
  {"-foreground"   , ".foreground"   , XrmoptionSepArg, "white"},
  {"-frequency"    , ".frequency"    , XrmoptionSepArg,      0 },
  {"-full-minute"  , ".full-minute"  , XrmoptionNoArg ,  "True"},
  {"-hide-index"   , ".hide-index"   , XrmoptionNoArg ,  "True"},
  {"-hide-mark"    , ".hide-mark"    , XrmoptionNoArg ,  "True"},
  {"-hide-second"  , ".hide-second"  , XrmoptionNoArg ,  "True"},
  {"-mono"         , ".mono"         , XrmoptionNoArg ,  "True"},
  {"-pin"          , ".pin"          , XrmoptionNoArg ,  "True"},
  {"-railway-clock", ".railway-clock", XrmoptionNoArg ,  "True"},
  {"-size"         , ".size"         , XrmoptionSepArg,      0 },
  {"-wireframe"    , ".wireframe"    , XrmoptionNoArg ,  "True"},
  {0, 0, 0, 0}
};

/* clock hand type definition */
typedef enum {
  htHOUR  ,
  htMINUTE,
  htSECOND,
  htMAX
} HAND_TYPE;

/* state definition */
typedef struct {
  Display *dpy               ;
  Window   window            ;
  XGWA     xgwa              ;
  Pixmap   pixmap            ;
  int      width             , /* [pixel] */
           height            ; /* [pixel] */
  double   radius            ;
  XPoint   center            ,
           orient            ;
  ulong    delay             , /* [usec] */
           last_msec         ; /* [msec] */
  double   theta[htMAX]      ; /* [degree] */

  /* face */
  int      face_type         ;
  char     face_name[MAX_OPT];

  /* commandline options */
  XColor   bg_color          ,
           fg_color          ;
  double   brightness        ;
  int      cycle             ,
           frequency         ;
  Bool     full_minute       ,
           hide_index        ,
           hide_mark         ,
           hide_second       ,
           mono              ,
           pin               ,
           railway_clock     ;
  double   size              ;
  Bool     wireframe         ;

  /* GC */
  GC       bg_gc             ,
           surface_gc        ,
           border_gc         ,
  /* packard specific */
           ring_sfc_gc       ,
           ring_bdr_gc       ,
           jewel_sfc_gc      ,
           jewel_bdr_gc      ,
  /* packard & ring specific */
           hand_sfc_gc[htMAX],
           hand_bdr_gc[htMAX],
  /* kclock & rwclock specific */
           second_sfc_gc     ,
           second_bdr_gc     ;

  /* watch specific */
  ushort   wch_high_color    ,
           wch_mid_color     ,
           wch_low_color     ,
           wch_color_step    ,
           wch_diff_colors[3];
  XColor   wch_color         ;
} state;

static void copy_xcolor (XColor *dest_color, XColor *source_color)
{
  dest_color->red   = source_color->red  ;
  dest_color->green = source_color->green;
  dest_color->blue  = source_color->blue ;
}

static void set_xcolor (XColor *color, ushort r, ushort g, ushort b)
{
  color->red   = r;
  color->green = g;
  color->blue  = b;
}

static void set_xcolor_rgb (XColor *color, RGB *rgb)
{
  color->red   = rgb->r;
  color->green = rgb->g;
  color->blue  = rgb->b;
}

static void adjust_brightness (state *st, XColor *color)
{
  set_xcolor (color,
              i_round (color->red   * st->brightness / 100.0),
              i_round (color->green * st->brightness / 100.0),
              i_round (color->blue  * st->brightness / 100.0));
}

static void set_bdr_xcolor (state *st, XColor *bdr_color, XColor *sfc_color)
{
  set_xcolor (bdr_color,
              average (st->bg_color.red  , sfc_color->red  ),
              average (st->bg_color.green, sfc_color->green),
              average (st->bg_color.blue , sfc_color->blue ));
}

static GC create_gc (state *st)
{
  return (XCreateGC (st->dpy, st->window, 0, NULL));
}

static void set_gc_xcolor (state *st, GC gc, XColor *color)
{
  XAllocColor (st->dpy, st->xgwa.colormap, color);
  XSetForeground (st->dpy, gc, color->pixel);
}


/*******************************/
/* Face definitions begin here */
/*******************************/


/* face types */
typedef enum {
  ftWCH,
  ftXCK,
  ftPKD,
  ftKCK,
  ftRWC,
  ftRNG,
  ftMAX
} FACE_TYPE;

/* face names */
static const char *face_names[ftMAX] = {
  "watch"  ,
  "xclock" ,
  "packard",
  "kclock" ,
  "rwclock",
  "ring"
};

/* face frequency defaults */
static const int face_frequency_defaults[ftMAX] = {
   5, /* watch   */
   1, /* xclock  */
   1, /* packard */
   1, /* kclock  */
  20, /* rwclock */
  10  /* ring    */
};

static void wch_draw_clock_pixmap (state *st);
static void xck_draw_clock_pixmap (state *st);
static void pkd_draw_clock_pixmap (state *st);
static void kck_draw_clock_pixmap (state *st);
static void rwc_draw_clock_pixmap (state *st);
static void rng_draw_clock_pixmap (state *st);
static void (*draw_clock_face_pixmap[ftMAX]) (state *st) = {
  wch_draw_clock_pixmap,
  xck_draw_clock_pixmap,
  pkd_draw_clock_pixmap,
  kck_draw_clock_pixmap,
  rwc_draw_clock_pixmap,
  rng_draw_clock_pixmap
};

static void wch_init (state *st);
static void xck_init (state *st);
static void pkd_init (state *st);
static void kck_init (state *st);
static void rwc_init (state *st);
static void rng_init (state *st);
static void (*face_init[ftMAX]) (state *st) = {
  wch_init,
  xck_init,
  pkd_init,
  kck_init,
  rwc_init,
  rng_init
};

static void wch_free (state *st);
static void xck_free (state *st);
static void pkd_free (state *st);
static void kck_free (state *st);
static void rwc_free (state *st);
static void rng_free (state *st);
static void (*face_free[ftMAX]) (state *st) = {
  wch_free,
  xck_free,
  pkd_free,
  kck_free,
  rwc_free,
  rng_free
};

static int face_type_of (const char* face_name)
{
  int ft, i;

  ft = -1;
  for (i = 0; i < ftMAX; i++)
    if (0 == strcmp (face_name, face_names[i])) {
      ft = i;
      break;
    }
  return (ft);
}

/* hand type specific definitions */
#define TRIANGLE_C   3
#define QUADRANGLE_C 4
typedef double DOUBLE_ARRAY[];
typedef double LINE_ARRAY[2]; /* outer or inner, inner or outer */
typedef double RECTANGLE_ARRAY[3]; /* outer or inner, inner or outer, width */
typedef double TRIANGLE_ARRAY[TRIANGLE_C];
typedef double QUADRANGLE_ARRAY[QUADRANGLE_C];
typedef double RING_ARRAY[2]; /* outer, inner */

static const int sgl_pi = 180 * 64;
static const int twi_pi = 360 * 64;

/* draw_hand functions */

static void draw_line_hand (state *st, LINE_ARRAY ratios, double theta, GC gc)
{
  double radii[2];
  double sin_theta, cos_theta;
  int    x1, y1, x2, y2;

  radii[0] = st->radius * ratios[0] / 100.0;
  radii[1] = st->radius * ratios[1] / 100.0;
  sin_theta = d_sin (theta);
  cos_theta = d_cos (theta);

  x1 = i_round (st->center.x + radii[0] * sin_theta);
  y1 = i_round (st->center.y - radii[0] * cos_theta);
  x2 = i_round (st->center.x + radii[1] * sin_theta);
  y2 = i_round (st->center.y - radii[1] * cos_theta);

  XDrawLine (st->dpy, st->pixmap, gc, x1, y1, x2, y2);
}

static void draw_polygon_hand (state *st, DOUBLE_ARRAY ratios,
  DOUBLE_ARRAY angles, int array_size, double theta, GC sfc_gc, GC bdr_gc,
  Bool edging)
{
  int     hv, array_width;
  double  hand_radius;
  XPoint *hand_vertexes;

  array_width = array_size + 1;
  hand_vertexes = (XPoint *) calloc (array_width, sizeof (XPoint));

  for (hv = 0; hv < array_size; hv++) {
    hand_radius = st->radius * ratios[hv] / 100.0;
    hand_vertexes[hv].x =
      i_round (st->center.x + hand_radius * d_sin (theta + angles[hv]));
    hand_vertexes[hv].y =
      i_round (st->center.y - hand_radius * d_cos (theta + angles[hv]));
  }
  hand_vertexes[array_size].x = hand_vertexes[0].x;
  hand_vertexes[array_size].y = hand_vertexes[0].y;

  XFillPolygon (st->dpy, st->pixmap, sfc_gc,
                hand_vertexes, array_width, Convex, CoordModeOrigin);
  if (edging)
    XDrawLines (st->dpy, st->pixmap, bdr_gc,
                hand_vertexes, array_width, CoordModeOrigin);

  free (hand_vertexes);
}

#define PREPARE_RECTANGLE_VARIABLES\
  XPoint *hand_vertexes;\
  double  apex, bottom, width;\
  double  cos_theta, sin_theta;\
  double  apex_cos, apex_sin, bottom_cos, bottom_sin, width_cos, width_sin;

#define SET_RECTANGLE_HAND_VERTEXES\
  hand_vertexes = (XPoint *) calloc (5, sizeof (XPoint));\
\
  apex       = st->radius * ratios[0] / 100.0;\
  bottom     = st->radius * ratios[1] / 100.0;\
  width      = st->radius * ratios[2] / 100.0;\
  cos_theta  = d_cos (theta);\
  sin_theta  = d_sin (theta);\
  apex_cos   = apex   * cos_theta;\
  apex_sin   = apex   * sin_theta;\
  bottom_cos = bottom * cos_theta;\
  bottom_sin = bottom * sin_theta;\
  width_cos  = width  * cos_theta;\
  width_sin  = width  * sin_theta;\
\
  hand_vertexes[0].x = i_round (st->center.x - width_cos + apex_sin  );\
  hand_vertexes[0].y = i_round (st->center.y - width_sin - apex_cos  );\
  hand_vertexes[1].x = i_round (st->center.x + width_cos + apex_sin  );\
  hand_vertexes[1].y = i_round (st->center.y + width_sin - apex_cos  );\
  hand_vertexes[2].x = i_round (st->center.x + width_cos + bottom_sin);\
  hand_vertexes[2].y = i_round (st->center.y + width_sin - bottom_cos);\
  hand_vertexes[3].x = i_round (st->center.x - width_cos + bottom_sin);\
  hand_vertexes[3].y = i_round (st->center.y - width_sin - bottom_cos);\
  hand_vertexes[4].x = hand_vertexes[0].x;\
  hand_vertexes[4].y = hand_vertexes[0].y;

static void draw_rectangle_hand (state *st, RECTANGLE_ARRAY ratios,
  double theta, GC sfc_gc, GC bdr_gc, Bool edging)
{
  PREPARE_RECTANGLE_VARIABLES

  SET_RECTANGLE_HAND_VERTEXES

  XFillPolygon (st->dpy, st->pixmap, sfc_gc,
                hand_vertexes, 5, Convex, CoordModeOrigin);
  if (edging)
    XDrawLines (st->dpy, st->pixmap, bdr_gc,
                hand_vertexes, 5, CoordModeOrigin);

  free (hand_vertexes);
}

static void draw_rounded_rectangle_hand (state *st, RECTANGLE_ARRAY ratios,
  double theta, GC sfc_gc, GC bdr_gc)
{
  PREPARE_RECTANGLE_VARIABLES
  int     x1, x2, y1, y2, wh, sa1, sa2;

  SET_RECTANGLE_HAND_VERTEXES

  x1  = st->center.x + i_round (apex   * sin_theta - width);
  y1  = st->center.y - i_round (apex   * cos_theta + width);
  x2  = st->center.x + i_round (bottom * sin_theta - width);
  y2  = st->center.y - i_round (bottom * cos_theta + width);
  wh  = i_round (width * 2.0);
  sa1 = i_round (-theta * 64.0);
  sa2 = i_round ((180.0 - theta) * 64.0);

  XFillPolygon (st->dpy, st->pixmap, sfc_gc,
                hand_vertexes, 5, Convex, CoordModeOrigin);
  XFillArc (st->dpy, st->pixmap, sfc_gc, x1, y1, wh, wh,  0, twi_pi);
  XFillArc (st->dpy, st->pixmap, sfc_gc, x2, y2, wh, wh,  0, twi_pi);

  XDrawLine (st->dpy, st->pixmap, bdr_gc,
             hand_vertexes[1].x, hand_vertexes[1].y,
             hand_vertexes[2].x, hand_vertexes[2].y);
  XDrawLine (st->dpy, st->pixmap, bdr_gc,
             hand_vertexes[3].x, hand_vertexes[3].y,
             hand_vertexes[4].x, hand_vertexes[4].y);
  XDrawArc (st->dpy, st->pixmap, bdr_gc, x1, y1, wh, wh, sa1, sgl_pi);
  XDrawArc (st->dpy, st->pixmap, bdr_gc, x2, y2, wh, wh, sa2, sgl_pi);

  free (hand_vertexes);
}

static void draw_circle (state *st, double center_ratio, double radius_ratio,
  double theta, GC sfc_gc, GC bdr_gc, Bool edging)
{
  double c, r;
  int    x, y, w;

  c = st->radius * center_ratio / 100.0;
  r = st->radius * radius_ratio / 100.0;
  x = st->center.x + i_round (c * d_sin (theta) - r);
  y = st->center.y - i_round (c * d_cos (theta) + r);
  w = i_round (r * 2.0);

  XFillArc (st->dpy, st->pixmap, sfc_gc, x, y, w, w, 0, twi_pi);
  if (edging)
    XDrawArc (st->dpy, st->pixmap, bdr_gc, x, y, w, w, 0, twi_pi);
}

static void draw_rounded_arc_hand (state *st, RING_ARRAY ratios, double theta,
  GC sfc_gc, GC bdr_gc, Bool hide_index)
{
  double ro, ri, c, r;
  int    a1, a2, ac1, ac2, xo, xi, xc1, xc2, yo, yi, yc1, yc2, wo, wi, wc;

  ro  = st->radius * ratios[0] / 100.0;
  ri  = st->radius * ratios[1] / 100.0;
  a1  = i_round ((90.0 - theta) * 64.0);
  a2  = i_round (theta * 64.0);
  xo  = st->center.x - i_round (ro);
  yo  = st->center.y - i_round (ro);
  xi  = st->center.x - i_round (ri);
  yi  = st->center.y - i_round (ri);
  wo  = i_round (ro * 2.0);
  wi  = i_round (ri * 2.0);
  c   = average (ro, ri);
  r   = (ro - ri) / 2.0;
  xc1 = st->center.x + i_round (-r);
  yc1 = st->center.y - i_round (c + r);
  xc2 = st->center.x + i_round (c * d_sin (theta) - r);
  yc2 = st->center.y - i_round (c * d_cos (theta) + r);
  wc  = i_round (ro - ri);
  ac1 = 90 * 64;
  ac2 = a1 + (180 * 64);

  XFillArc (st->dpy, st->pixmap, sfc_gc,    xo , yo , wo, wo, a1 , a2    );
  XFillArc (st->dpy, st->pixmap, sfc_gc,    xc1, yc1, wc, wc, 0.0, twi_pi);
  XFillArc (st->dpy, st->pixmap, sfc_gc,    xc2, yc2, wc, wc, 0.0, twi_pi);
  XFillArc (st->dpy, st->pixmap, st->bg_gc, xi , yi , wi, wi, 0.0, twi_pi);
  XDrawArc (st->dpy, st->pixmap, bdr_gc,    xo , yo , wo, wo, a1 , a2    );
  XDrawArc (st->dpy, st->pixmap, bdr_gc,    xi , yi , wi, wi, a1 , a2    );
  XDrawArc (st->dpy, st->pixmap, bdr_gc,    xc1, yc1, wc, wc, ac1, sgl_pi);
  XDrawArc (st->dpy, st->pixmap, bdr_gc,    xc2, yc2, wc, wc, ac2, sgl_pi);

  if (!hide_index) {
    int minute;

    for (minute = 0; minute < 60; minute += 5) {
      double minx6;

      minx6 = minute * 6.0;
      if (theta < minx6)
        break;
      draw_line_hand (st, ratios, minx6, bdr_gc);
    }
  }
}


/*******************************************************/
/* Watch face                                          */
/* Original anclock face based on my favorite watch.   */
/* Copyright(c) 2009-2022 PT2K <webmaster@pt2k.xii.jp> */
/*******************************************************/

/* color range settings */
static const ushort wch_high_color_default = 0xF000;
static const ushort wch_mid_color_default  = 0xE000;
static const ushort wch_low_color_default  = 0xD000;
static const ushort wch_color_step_default = 0x0010;

/* ratio definitions [%] */
static LINE_ARRAY wch_mark_ratios = {
   100.0,  98.0 /* outer, inner */
};
static RECTANGLE_ARRAY wch_index_ratios = {
    96.0,  88.0,   0.7 /* outer, inner, width */
};
static QUADRANGLE_ARRAY wch_hand_radius_ratios[htMAX] = {
  { 65.0,   4.0,  15.0,   4.0}, /* Hour */
  { 95.0,   3.0,  15.0,   3.0}, /* Min  */
  { 98.0,  29.0,  30.0,  29.0}  /* Sec  */
};
static const double wch_center_circle_ratio  =  2.0;

/* angle definitions [degree] */
static QUADRANGLE_ARRAY wch_hand_vertex_angles[htMAX] = {
  {  0.0,  54.0, 180.0, 306.0}, /* Hour */
  {  0.0,  54.0, 180.0, 306.0}, /* Min  */
  {  0.0, 178.0, 180.0, 182.0}  /* Sec  */
};

static void set_wch_color (state *st, XColor *color)
{
  XColor bdr_color;

  if (st->wireframe)
    set_gc_xcolor (st, st->border_gc , color);
  else {
    set_gc_xcolor (st, st->surface_gc, color);
    set_bdr_xcolor (st, &bdr_color, color);
    set_gc_xcolor (st, st->border_gc, &bdr_color);
  }

  copy_xcolor (&st->wch_color, color);
}

static void wch_change_color (state *st)
{
  ushort temp_colors[3], buff;
  XColor new_color;
  int    color_orient;
  int    i, j;

  if (st->wch_mid_color == st->wch_color.red) {
    color_orient = i_random (2) * 2 - 1;
    for (i = 0; i < 3; i++)
      st->wch_diff_colors[i] = st->wch_color_step * color_orient;
    st->wch_diff_colors[i_random (3)] = -2 * st->wch_color_step * color_orient;
  }

  temp_colors[0] = st->wch_color.red  ;
  temp_colors[1] = st->wch_color.green;
  temp_colors[2] = st->wch_color.blue ;
  for (i = 0; i < 3; i++) {
    buff = temp_colors[i] + st->wch_diff_colors[i];
    if ((buff < st->wch_low_color) || (st->wch_high_color < buff)) {
      for (j = 0; j < 3; j++)
        st->wch_diff_colors[j] = -st->wch_diff_colors[j];
      break;
    }
  }

  for (i = 0; i < 3; i++)
    temp_colors[i] += st->wch_diff_colors[i];

  set_xcolor (&new_color, temp_colors[0], temp_colors[1], temp_colors[2]);
  set_wch_color (st, &new_color);
}

static void wch_draw_marks (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute++)
    draw_line_hand (st, wch_mark_ratios, minute * 6.0,
                    st->wireframe ? st->border_gc : st->surface_gc);
}

static void wch_draw_indexes (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute += 5)
    draw_rectangle_hand (st, wch_index_ratios, minute * 6.0,
                         st->surface_gc, st->border_gc, True);
}

static void wch_draw_hand (state *st, HAND_TYPE ht)
{
  draw_polygon_hand (st, wch_hand_radius_ratios[ht], wch_hand_vertex_angles[ht],
                     QUADRANGLE_C, st->theta[ht],
                     st->surface_gc, st->border_gc, True);
}

static void wch_draw_center_circle (state *st)
{
  draw_circle (st, 0.0, wch_center_circle_ratio, 0.0,
               st->surface_gc, st->border_gc, True);
}

static void wch_draw_clock_pixmap (state *st)
{
  if (!st->mono)
    wch_change_color(st);
  if (!st->hide_mark)
    wch_draw_marks (st);
  if (!st->hide_index)
    wch_draw_indexes (st);
  wch_draw_hand (st, htHOUR);
  wch_draw_hand (st, htMINUTE);
  if (!st->hide_second)
    wch_draw_hand (st, htSECOND);
  wch_draw_center_circle (st);
}

static void wch_init (state *st)
{
  XColor sfc_color, bdr_color, mid_color;

  st->surface_gc = create_gc (st);
  st->border_gc  = create_gc (st);

  if (st->mono) {
    if (st->wireframe) {
      copy_xcolor (&sfc_color, &st->bg_color);
      copy_xcolor (&bdr_color, &st->fg_color);
    }
    else {
      copy_xcolor (&sfc_color, &st->fg_color);
      set_bdr_xcolor (st, &bdr_color, &sfc_color);
    }
    set_gc_xcolor (st, st->surface_gc, &sfc_color);
    set_gc_xcolor (st, st->border_gc , &bdr_color);
  }
  else {
    st->wch_high_color =
      i_round (wch_high_color_default * st->brightness / 100.0);
    st->wch_mid_color  =
      i_round (wch_mid_color_default  * st->brightness / 100.0);
    st->wch_low_color  =
      i_round (wch_low_color_default  * st->brightness / 100.0);
    st->wch_color_step =
      i_round (wch_color_step_default * st->brightness / 100.0);
    set_xcolor (&mid_color,
                st->wch_mid_color, st->wch_mid_color, st->wch_mid_color);
    set_wch_color (st, &mid_color);
  }
}

static void wch_free (state *st)
{
  XFreeGC (st->dpy, st->surface_gc);
  XFreeGC (st->dpy, st->border_gc );
}


/**************************************************************/
/* XCLOCK face                                                */
/* Emulation of XCLOCK.                                       */
/* Copyright 1989, 1998  The Open Group.                      */
/*   Hacked from Tony Della Fera's much hacked clock program. */
/* Coded by PT2K.                                             */
/**************************************************************/

/* ratio definitions [%] */
static RECTANGLE_ARRAY xck_mark_ratios = {
  100.0,  95.0,   0.4 /* outer, inner, width */
};
static RECTANGLE_ARRAY xck_index_ratios = {
  100.0,  90.0,   0.4 /* outer, inner, width */
};

/* angle definitions [degree] */
static TRIANGLE_ARRAY xck_hour_hand_vertex_angles = {
    0.0, 135.0, 225.0
};
static TRIANGLE_ARRAY xck_minute_hand_vertex_angles = {
    0.0, 135.0, 225.0
};

/* ratio definitions [%] */
static TRIANGLE_ARRAY xck_hour_hand_radius_ratios = {
   40.0,   9.0,   9.0
};
static TRIANGLE_ARRAY xck_minute_hand_radius_ratios = {
   70.0,   9.0,   9.0
};

/* second hand definitions [%] */
static RECTANGLE_ARRAY xck_second_hand_ratios = {
   95.0,   0.0,   0.9 /* outer, inner, width */
};
static QUADRANGLE_ARRAY xck_jewel_ratios = {
   93.0,  81.0,  69.0,  81.0
};
static QUADRANGLE_ARRAY xck_jewel_angles = {
    0.0,   3.1,   0.0,  -3.1
};

static void xck_draw_marks (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute++)
    draw_rectangle_hand (st, xck_mark_ratios, minute * 6.0,
                         st->surface_gc, st->border_gc, True);
}

static void xck_draw_indexes (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute += 5)
    draw_rectangle_hand (st, xck_index_ratios, minute * 6.0,
                         st->surface_gc, st->border_gc, True);
}

static void xck_draw_clock_pixmap (state *st)
{
  if (!st->hide_mark)
    xck_draw_marks (st);
  if (!st->hide_index)
    xck_draw_indexes (st);

  /* hour */
  draw_polygon_hand (st,
                     xck_hour_hand_radius_ratios,
                     xck_hour_hand_vertex_angles,
                     TRIANGLE_C, st->theta[htHOUR],
                     st->surface_gc, st->border_gc, st->wireframe);
  /* minute */
  draw_polygon_hand (st,
                     xck_minute_hand_radius_ratios,
                     xck_minute_hand_vertex_angles,
                     TRIANGLE_C, st->theta[htMINUTE],
                     st->surface_gc, st->border_gc, st->wireframe);
  /* second */
  if (!st->hide_second) {
    draw_rectangle_hand (st, xck_second_hand_ratios, st->theta[htSECOND],
                         st->surface_gc, st->border_gc, st->wireframe);
    /* jewel */
    draw_polygon_hand (st, xck_jewel_ratios, xck_jewel_angles,
                       QUADRANGLE_C, st->theta[htSECOND],
                       st->surface_gc, st->border_gc, st->wireframe);
  }
}

static void xck_init (state *st)
{
  XColor sfc_color, bdr_color;

  st->surface_gc = create_gc (st);
  st->border_gc  = create_gc (st);

  if (st->wireframe) {
    copy_xcolor (&sfc_color, &st->bg_color);
    copy_xcolor (&bdr_color, &st->fg_color);
    if (!st->mono)
      adjust_brightness (st, &bdr_color);
  }
  else {
    copy_xcolor (&sfc_color, &st->fg_color);
    if (!st->mono)
      adjust_brightness (st, &sfc_color);
    set_bdr_xcolor (st, &bdr_color, &sfc_color);
  }
  set_gc_xcolor (st, st->surface_gc, &sfc_color);
  set_gc_xcolor (st, st->border_gc , &bdr_color);
}

static void xck_free (state *st)
{
  XFreeGC (st->dpy, st->surface_gc);
  XFreeGC (st->dpy, st->border_gc );
}


/********************************************************************/
/* Packard face                                                     */
/* Emulation of Packard's clock from xclockmore.                    */
/* Copyright(c) 1995, 2014 by Jeremie PETIT <petit@aurora.unice.fr> */
/* Coded by PT2K.                                                   */
/********************************************************************/

/* ratio definitions [%] */
static const double pkd_ring_outer_ratio   = 100.0;
static const double pkd_ring_inner_ratio   =  95.0;
static const double pkd_jewel_center_ratio =  90.0;
static const double pkd_jewel_radius_ratio =   5.0;

static RECTANGLE_ARRAY pkd_hand_vertex_ratios[htMAX] = {
  { 50.0,   0.0,   5.0}, /* Hour */
  { 75.0,   0.0,   2.5}, /*  Min */
  { 80.0,   0.0,   1.0}  /*  Sec */
};

#define COLOR_LIST_COUNT 64
static RGB color_list[COLOR_LIST_COUNT]= {
  {0x0000, 0x0F00, 0xFF00}, {0x0000, 0x2700, 0xFF00}, {0x0000, 0x3F00, 0xFF00},
  {0x0000, 0x5700, 0xFF00}, {0x0000, 0x6F00, 0xFF00}, {0x0000, 0x8700, 0xFF00},
  {0x0000, 0x9F00, 0xFF00}, {0x0000, 0xB700, 0xFF00}, {0x0000, 0xCF00, 0xFF00},
  {0x0000, 0xE700, 0xFF00}, {0x0000, 0xFF00, 0x0F00}, {0x0000, 0xFF00, 0x2700},
  {0x0000, 0xFF00, 0x3F00}, {0x0000, 0xFF00, 0x5700}, {0x0000, 0xFF00, 0x6F00},
  {0x0000, 0xFF00, 0x8700}, {0x0000, 0xFF00, 0x9F00}, {0x0000, 0xFF00, 0xB700},
  {0x0000, 0xFF00, 0xCF00}, {0x0000, 0xFF00, 0xE700}, {0x0000, 0xFF00, 0xFF00},
  {0x0700, 0x0000, 0xFF00}, {0x0700, 0xFF00, 0x0000}, {0x1F00, 0x0000, 0xFF00},
  {0x1F00, 0xFF00, 0x0000}, {0x3700, 0x0000, 0xFF00}, {0x3700, 0xFF00, 0x0000},
  {0x4F00, 0x0000, 0xFF00}, {0x4F00, 0xFF00, 0x0000}, {0x6700, 0x0000, 0xFF00},
  {0x6700, 0xFF00, 0x0000}, {0x7F00, 0x0000, 0xFF00}, {0x7F00, 0xFF00, 0x0000},
  {0x9700, 0x0000, 0xFF00}, {0x9700, 0xFF00, 0x0000}, {0xAF00, 0x0000, 0xFF00},
  {0xAF00, 0xFF00, 0x0000}, {0xC700, 0x0000, 0xFF00}, {0xC700, 0xFF00, 0x0000},
  {0xDF00, 0x0000, 0xFF00}, {0xDF00, 0xFF00, 0x0000}, {0xF700, 0x0000, 0xFF00},
  {0xF700, 0xFF00, 0x0000}, {0xFF00, 0x0000, 0x0000}, {0xFF00, 0x0000, 0x1700},
  {0xFF00, 0x0000, 0x2F00}, {0xFF00, 0x0000, 0x4700}, {0xFF00, 0x0000, 0x5F00},
  {0xFF00, 0x0000, 0x7700}, {0xFF00, 0x0000, 0x8F00}, {0xFF00, 0x0000, 0xA700},
  {0xFF00, 0x0000, 0xBF00}, {0xFF00, 0x0000, 0xD700}, {0xFF00, 0x0000, 0xEF00},
  {0xFF00, 0x1700, 0x0000}, {0xFF00, 0x2F00, 0x0000}, {0xFF00, 0x4700, 0x0000},
  {0xFF00, 0x5F00, 0x0000}, {0xFF00, 0x7700, 0x0000}, {0xFF00, 0x8F00, 0x0000},
  {0xFF00, 0xA700, 0x0000}, {0xFF00, 0xBF00, 0x0000}, {0xFF00, 0xD700, 0x0000},
  {0xFF00, 0xEF00, 0x0000}
};

static void pkd_get_random_color (state *st, GC sfc_gc, GC bdr_gc)
{
  XColor color, bdr_color;

  set_xcolor_rgb (&color, &color_list[i_random (COLOR_LIST_COUNT)]);
  if (!st->mono)
    adjust_brightness (st, &color);
  if (st->wireframe)
    set_gc_xcolor  (st,  bdr_gc   , &color    );
  else {
    set_gc_xcolor  (st,  sfc_gc   , &color    );
    set_bdr_xcolor (st, &bdr_color, &color    );
    set_gc_xcolor  (st,  bdr_gc   , &bdr_color);
  }
}

static void pkd_change_color (state *st)
{
  int ht;

  pkd_get_random_color (st, st->ring_sfc_gc , st->ring_bdr_gc );
  pkd_get_random_color (st, st->jewel_sfc_gc, st->jewel_bdr_gc);
  for (ht = 0; ht < htMAX; ht++) {
    pkd_get_random_color (st, st->hand_sfc_gc[ht], st->hand_bdr_gc[ht]);
  }
}

static void pkd_draw_ring (state *st)
{
  draw_circle (st, 0.0, pkd_ring_outer_ratio, 0.0,
               st->ring_sfc_gc, st->ring_bdr_gc, True);
  draw_circle (st, 0.0, pkd_ring_inner_ratio, 0.0,
               st->bg_gc, st->ring_bdr_gc, True);
  /* jewel */
  draw_circle (st, pkd_jewel_center_ratio, pkd_jewel_radius_ratio, 0.0,
               st->jewel_sfc_gc, st->jewel_bdr_gc, True);
}

static void pkd_draw_hand (state *st, HAND_TYPE ht)
{
  draw_rounded_rectangle_hand (st, pkd_hand_vertex_ratios[ht], st->theta[ht],
                               st->hand_sfc_gc[ht], st->hand_bdr_gc[ht]);
}

static void pkd_draw_clock_pixmap (state *st)
{
  pkd_draw_ring (st);
  pkd_draw_hand (st, htHOUR);
  pkd_draw_hand (st, htMINUTE);
  if (!st->hide_second)
    pkd_draw_hand (st, htSECOND);
}

static void pkd_init (state *st)
{
  int ht;

  st->ring_sfc_gc  = create_gc (st);
  st->ring_bdr_gc  = create_gc (st);
  st->jewel_sfc_gc = create_gc (st);
  st->jewel_bdr_gc = create_gc (st);
  for (ht = 0; ht < htMAX; ht++) {
    st->hand_sfc_gc[ht] = create_gc (st);
    st->hand_bdr_gc[ht] = create_gc (st);
  }

  if (st->wireframe) {
    set_gc_xcolor (st, st->ring_sfc_gc , &st->bg_color);
    set_gc_xcolor (st, st->jewel_sfc_gc, &st->bg_color);
    for (ht = 0; ht < htMAX; ht++)
      set_gc_xcolor (st, st->hand_sfc_gc[ht], &st->bg_color);
  }

  pkd_change_color (st);
}

static void pkd_free (state *st)
{
  int ht;

  XFreeGC (st->dpy, st->ring_sfc_gc );
  XFreeGC (st->dpy, st->ring_bdr_gc );
  XFreeGC (st->dpy, st->jewel_sfc_gc);
  XFreeGC (st->dpy, st->jewel_bdr_gc);
  for (ht = 0; ht < htMAX; ht++) {
    XFreeGC (st->dpy, st->hand_sfc_gc[ht]);
    XFreeGC (st->dpy, st->hand_bdr_gc[ht]);
  }
}


/********************************************************/
/* KCLOCK face                                          */
/* Resembles KDE screensaver clock.                     */
/* Copyright(c) 2022 Manfred Schwarb <manfred99@gmx.ch> */
/********************************************************/

/* ratio definitions [%] */
static RECTANGLE_ARRAY kck_mark_ratios = {
    100.0,  94.0,   0.8   /* outer, inner, width */
};
static RECTANGLE_ARRAY kck_index_ratios = {
    100.0,  84.0,   2.0   /* outer, inner, width */
};
static RECTANGLE_ARRAY kck_hand_radius_ratios[htMAX] = {
  {  60.0,   0.0,   3.0}, /* Hour */
  {  92.0,   0.0,   2.0}, /* Min  */
  {  92.0,   0.0,   1.4}  /* Sec  */
};

static const double kck_center_circle_ratio[htMAX] = {
      5.2,   5.2,   3.8   /* Hour, Min, Sec */
};

static void kck_draw_marks (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute++)
    draw_rectangle_hand (st, kck_mark_ratios, minute * 6.0,
                         st->surface_gc, st->border_gc, True);
}

static void kck_draw_indexes (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute += 5)
    draw_rectangle_hand (st, kck_index_ratios, minute * 6.0,
                         st->surface_gc, st->border_gc, True);
}

static void kck_draw_hand (state *st, HAND_TYPE ht)
{
  GC hand_sfc_gc, hand_bdr_gc;

  hand_sfc_gc = (htSECOND == ht) ? st->second_sfc_gc : st->surface_gc;
  hand_bdr_gc = (htSECOND == ht) ? st->second_bdr_gc : st->border_gc ;
  draw_rectangle_hand (st, kck_hand_radius_ratios[ht], st->theta[ht],
                       hand_sfc_gc, hand_bdr_gc, st->wireframe);
}

static void kck_draw_center_circle (state *st, HAND_TYPE ht)
{
  GC hand_sfc_gc, hand_bdr_gc;

  hand_sfc_gc = (htSECOND == ht) ? st->second_sfc_gc : st->surface_gc;
  hand_bdr_gc = (htSECOND == ht) ? st->second_bdr_gc : st->border_gc ;
  draw_circle (st, 0.0, kck_center_circle_ratio[ht], 0.0,
               hand_sfc_gc, hand_bdr_gc, st->wireframe);
}

static void kck_draw_clock_pixmap (state *st)
{
  if (!st->hide_mark)
    kck_draw_marks (st);
  if (!st->hide_index)
    kck_draw_indexes (st);
  kck_draw_hand (st, htHOUR);
  kck_draw_center_circle (st, htHOUR);
  kck_draw_hand (st, htMINUTE);
  kck_draw_center_circle (st, htMINUTE);
  if (!st->hide_second) {
    kck_draw_hand (st, htSECOND);
    kck_draw_center_circle (st, htSECOND);
  }
}

static void kck_init (state *st)
{
  XColor sec_hnd_color;

  xck_init (st);
  st->second_sfc_gc = create_gc (st);
  st->second_bdr_gc = create_gc (st);

  set_xcolor (&sec_hnd_color, 0xFFFF, 0, 0);
  if (!st->mono)
    adjust_brightness (st, &sec_hnd_color);

  if (st->wireframe) {
    set_gc_xcolor (st, st->second_sfc_gc, &st->bg_color );
    set_gc_xcolor (st, st->second_bdr_gc, &sec_hnd_color);
  }
  else {
    set_gc_xcolor (st, st->second_sfc_gc, &sec_hnd_color);
    sec_hnd_color.red = average (st->bg_color.red, sec_hnd_color.red);
    set_gc_xcolor (st, st->second_bdr_gc, &sec_hnd_color);
  }

  st->full_minute = True;
}

static void kck_free (state *st)
{
  xck_free (st);
  XFreeGC (st->dpy, st->second_sfc_gc);
  XFreeGC (st->dpy, st->second_bdr_gc);
}


/**************************************************/
/* RWClock face                                   */
/* Emulation of Swiss railway clock.              */
/* Copyright(c) 2022 PT2K <webmaster@pt2k.xii.jp> */
/**************************************************/

/* ratio definitions [%] */
static RECTANGLE_ARRAY rwc_mark_ratios = {
  100.0,  93.6,   1.3  /* outer, inner, width */
};
static RECTANGLE_ARRAY rwc_index_ratios = {
  100.0,  76.6,   3.5  /* outer, inner, width */
};
static QUADRANGLE_ARRAY rwc_hour_hand_radius_ratio = {
   69.7,  69.7,  25.0,  25.0
};
static QUADRANGLE_ARRAY rwc_minute_hand_radius_ratio = {
   94.9,  94.9,  24.0,  24.0
};
static RECTANGLE_ARRAY rwc_second_hand_radius_ratio = {
   66.8, -34.0,   1.4  /* outer, inner, width */
};
static double rwc_jewel_center_ratio = 66.8;
static double rwc_jewel_radius_ratio =  9.8;

/* angle definitions [degree] */
static QUADRANGLE_ARRAY rwc_hour_hand_vertex_angle = {
  356.2,   3.8, 165.0, 195.0
};
static QUADRANGLE_ARRAY rwc_minute_hand_vertex_angle = {
  357.7,   2.3, 167.4, 192.6
};

static void rwc_draw_marks (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute++)
    draw_rectangle_hand (st, rwc_mark_ratios, minute * 6.0,
                         st->surface_gc, st->border_gc, True);
}

static void rwc_draw_indexes (state *st)
{
  int minute;

  for (minute = 0; minute < 60; minute += 5)
    draw_rectangle_hand (st, rwc_index_ratios, minute * 6.0,
                         st->surface_gc, st->border_gc, True);
}

static void rwc_draw_hand (state *st, HAND_TYPE ht)
{
  GC hand_sfc_gc, hand_bdr_gc;

  hand_sfc_gc = (htSECOND == ht) ? st->second_sfc_gc : st->surface_gc;
  hand_bdr_gc = (htSECOND == ht) ? st->second_bdr_gc : st->border_gc ;
  if (htHOUR == ht)
    draw_polygon_hand (st, rwc_hour_hand_radius_ratio,
                       rwc_hour_hand_vertex_angle, QUADRANGLE_C,
                       st->theta[ht], st->surface_gc, st->border_gc, True);
  if (htMINUTE == ht)
    draw_polygon_hand (st, rwc_minute_hand_radius_ratio,
                       rwc_minute_hand_vertex_angle, QUADRANGLE_C,
                       st->theta[ht], st->surface_gc, st->border_gc, True);
  if (htSECOND == ht) {
    draw_rectangle_hand (st, rwc_second_hand_radius_ratio, st->theta[ht],
                         hand_sfc_gc, hand_bdr_gc, st->wireframe);
    /* jewel */
    draw_circle (st, rwc_jewel_center_ratio, rwc_jewel_radius_ratio,
                 st->theta[ht], hand_sfc_gc, hand_bdr_gc, st->wireframe);
  }
}

static void rwc_draw_clock_pixmap (state *st)
{
  if (!st->hide_mark)
    rwc_draw_marks (st);
  if (!st->hide_index)
    rwc_draw_indexes (st);
  rwc_draw_hand (st, htHOUR);
  rwc_draw_hand (st, htMINUTE);
  if (!st->hide_second) {
    rwc_draw_hand (st, htSECOND);
  }
}

static void rwc_init (state *st)
{
  kck_init (st);
  st->railway_clock = True;
}

static void rwc_free (state *st)
{
  kck_free (st);
}


/**************************************************/
/* Ring face                                      */
/* Copyright(c) 2022 PT2K <webmaster@pt2k.xii.jp> */
/**************************************************/

/* ratio definitions [%] */
static RING_ARRAY rng_hand_ratios[htMAX] = {
  { 80.0,  75.0}, /* Hour */
  { 90.0,  85.0}, /*  Min */
  {100.0,  95.0}  /*  Sec */
};

static void rng_change_color (state *st)
{
  int ht;

  if (!st->mono)
    for (ht = 0; ht < htMAX; ht++)
      pkd_get_random_color (st, st->hand_sfc_gc[ht], st->hand_bdr_gc[ht]);
}

static void rng_draw_hand (state *st, HAND_TYPE ht)
{
  draw_rounded_arc_hand (st, rng_hand_ratios[ht], st->theta[ht],
                         st->hand_sfc_gc[ht], st->hand_bdr_gc[ht],
                         st->hide_index);
}

static void rng_draw_clock_pixmap (state *st)
{
  if (!st->hide_second)
    rng_draw_hand (st, htSECOND);
  rng_draw_hand (st, htMINUTE);
  rng_draw_hand (st, htHOUR);
}

static void rng_init (state *st)
{
  int ht;

  for (ht = 0; ht < htMAX; ht++) {
    st->hand_sfc_gc[ht] = create_gc (st);
    st->hand_bdr_gc[ht] = create_gc (st);
    if (st->wireframe) {
      set_gc_xcolor (st, st->hand_sfc_gc[ht], &st->bg_color);
      set_gc_xcolor (st, st->hand_bdr_gc[ht], &st->fg_color);
    }
    else {
      set_gc_xcolor (st, st->hand_sfc_gc[ht], &st->fg_color);
      set_gc_xcolor (st, st->hand_bdr_gc[ht], &st->bg_color);
    }
  }

  rng_change_color (st);
}

static void rng_free (state *st)
{
  int ht;

  for (ht = 0; ht < htMAX; ht++) {
    XFreeGC (st->dpy, st->hand_sfc_gc[ht]);
    XFreeGC (st->dpy, st->hand_bdr_gc[ht]);
  }
}


/*****************************/
/* Face definitions end here */
/*****************************/


static void create_pixmap (state *st)
{
  if (st->pixmap)
    XFreePixmap (st->dpy, st->pixmap);
  st->pixmap = XCreatePixmap (st->dpy, st->window, st->width, st->height,
                              st->xgwa.depth);
}

static void move_clock (state *st)
{
  if (st->center.x - st->radius <= 0)
    st->orient.x =  1;
  if (st->width <= st->center.x + st->radius)
    st->orient.x = -1;
  if (st->center.y - st->radius <= 0)
    st->orient.y =  1;
  if (st->height <= st->center.y + st->radius)
    st->orient.y = -1;
  st->center.x += st->orient.x;
  st->center.y += st->orient.y;
}

static void draw_clock_pixmap (state *st)
{
  static const double rwc_sec_stop     = 1000.0 * 60.0 * (58.5 / 60.0);
  static const double rwc_circle_ratio = 60.0 / 58.5;

  struct timeval  tv;
  struct tm      *now;
  ulong           msec;
  double          numerator, denominator;

  gettimeofday (&tv, NULL);
  now = localtime (&tv.tv_sec);
  msec = (tv.tv_usec - tv.tv_usec % st->delay) / 1000;

  numerator               = 1000.0 * now->tm_sec + msec;
  denominator             = 60.0 * 1000.0;
  st->theta[htSECOND]     = 360.0 * numerator / denominator;
  if (st->railway_clock) {
    if (numerator < rwc_sec_stop)
      st->theta[htSECOND] = st->theta[htSECOND] * rwc_circle_ratio;
    else
      st->theta[htSECOND] = 0.0;
  }
  numerator               = 60000.0 * now->tm_min + numerator;
  denominator             = 60.0 * denominator;
  if (st->full_minute)
    st->theta[htMINUTE]   = 360.0 * 60000.0 * now->tm_min / denominator;
  else
    st->theta[htMINUTE]   = 360.0 * numerator / denominator;
  numerator               = 3600000.0 * (now->tm_hour % 12) + numerator;
  denominator             = 12.0 * denominator;
  st->theta[htHOUR]       = 360.0 * numerator / denominator;

  if ((1 == st->frequency) || (msec < st->last_msec)) {
    if ((0 != st->cycle) && (0 == now->tm_sec % st->cycle)) {
      if (ftPKD == st->face_type)
        pkd_change_color (st);
      if (ftRNG == st->face_type)
        rng_change_color (st);
    }
    if (!st->pin)
      move_clock (st);
  }
  st->last_msec = msec;

  XFillRectangle (st->dpy, st->pixmap, st->bg_gc, 0, 0, st->width, st->height);
  draw_clock_face_pixmap[st->face_type] (st);
}

static void option_warning (char *opt, char *val, char *msg, char *newval)
{
  fprintf (stderr, "Warning: %s %s is %s; using %s.\n", opt, val, msg, newval);
}

static void correct_options (state *st)
{
  static char *msg_invalid    = "invalid"     ;
  static char *msg_range      = "out of range";
  static char *msg_odd        = "odd"         ;
  static char *fmt_brightness = "%.1f"        ;
  static char *fmt_cycle      = "%d"          ;
  static char *fmt_frequency  = "%d"          ;
  static char *fmt_size       = "%.1f"        ;
  static const double  min_brightness     =  10.0;
  static const double  default_brightness =  70.0;
  static const double  max_brightness     = 100.0;
  static const int     min_cycle          =   0  ;
  static const int     default_cycle      =  20  ;
  static const int     max_cycle          =  60  ;
  static const int     min_frequency      =   0  ;
  static const int     max_frequency      =  25  ;
  static const double  min_size           =  10.0;
  static const double  default_size       =  85.0;
  static const double  max_size           =  99.0;

  char val[MAX_OPT], newval[MAX_OPT];
  int  ft;

  st->face_type = face_type_of (st->face_name);
  if (st->face_type < 0) {
    strcpy (val   , st->face_name    );
    strcpy (newval, face_names[ftWCH]);
    option_warning ("face", val, msg_invalid, newval);
    fprintf (stderr, "Notice: available face names are ");
    for (ft = 0; ft < ftMAX - 1; ft++)
      fprintf (stderr, "%s, ", face_names[ft]);
    fprintf (stderr, "and %s.\n", face_names[ftMAX - 1]);
    st->face_type = ftWCH;
    strcpy (st->face_name, face_names[ftWCH]);
  }

  if ((st->brightness < min_brightness) || (max_brightness < st->brightness)) {
    sprintf (val   , fmt_brightness, st->brightness    );
    sprintf (newval, fmt_brightness, default_brightness);
    option_warning ("brightness", val, msg_range, newval);
    st->brightness = default_brightness;
  }

  if ((st->cycle < min_cycle) || (max_cycle < st->cycle)) {
    sprintf (val   , fmt_cycle, st->cycle    );
    sprintf (newval, fmt_cycle, default_cycle);
    option_warning ("cycle", val, msg_range, newval);
    st->cycle = default_cycle;
  }
  if ((0 != st->cycle) && (0 != 60 % st->cycle)) {
    sprintf (val   , fmt_cycle, st->cycle    );
    sprintf (newval, fmt_cycle, default_cycle);
    option_warning ("cycle", val, msg_odd, newval);
    st->cycle = default_cycle;
  }

  if ((st->frequency < min_frequency) || (max_frequency < st->frequency)) {
    sprintf (val   , fmt_frequency, st->frequency);
    sprintf (newval, fmt_frequency, 0            );
    option_warning ("frequency", val, msg_range, newval);
    st->frequency = 0;
  }
  if ((0 != st->frequency) && (0 != 1000 % st->frequency)) {
    sprintf (val   , fmt_frequency, st->frequency);
    sprintf (newval, fmt_frequency, 0            );
    option_warning ("frequency", val, msg_odd, newval);
    st->frequency = 0;
  }
  if (0 == st->frequency)
    st->frequency = face_frequency_defaults[st->face_type];

  if ((st->size < min_size) || (max_size < st->size)) {
    sprintf (val   , fmt_size, st->size    );
    sprintf (newval, fmt_size, default_size);
    option_warning ("size", val, msg_range, newval);
    st->size = default_size;
  }

  if (st->railway_clock)
    st->full_minute = True;
}

static void *anclock_init (Display *dpy, Window window)
{
  state *st;
  char  *face_name;
  ulong  pixel;

  st                = (state *) calloc (1, sizeof (*st));
  st->dpy           = dpy;
  st->window        = window;
  XGetWindowAttributes (st->dpy, st->window, &st->xgwa);
  st->width         = st->xgwa.width;
  st->height        = st->xgwa.height;

  st->face_type     = ftWCH;
  face_name         = get_string_resource  (dpy, "face"         , "Face"   );
  strcpy (st->face_name, face_name);
  free (face_name);
  st->brightness    = get_float_resource   (dpy, "brightness"   , "Float"  );
  st->cycle         = get_integer_resource (dpy, "cycle"        , "Integer");
  st->frequency     = get_integer_resource (dpy, "frequency"    , "Integer");
  st->full_minute   = get_boolean_resource (dpy, "full-minute"  , "Boolean");
  st->hide_index    = get_boolean_resource (dpy, "hide-index"   , "Boolean");
  st->hide_mark     = get_boolean_resource (dpy, "hide-mark"    , "Boolean");
  st->hide_second   = get_boolean_resource (dpy, "hide-second"  , "Boolean");
  st->mono          = get_boolean_resource (dpy, "mono"         , "Boolean");
  st->pin           = get_boolean_resource (dpy, "pin"          , "Boolean");
  st->railway_clock = get_boolean_resource (dpy, "railway-clock", "Boolean");
  st->size          = get_float_resource   (dpy, "size"         , "Float"  );
  st->wireframe     = get_boolean_resource (dpy, "wireframe"    , "Boolean");

  correct_options (st);

  st->delay = 1000000 / st->frequency;
  st->last_msec = 0;

  st->bg_gc = create_gc (st);

  pixel = get_pixel_resource (dpy, st->xgwa.colormap,
                              "background", "Background");
  st->bg_color.pixel = pixel;
  XQueryColor (dpy, st->xgwa.colormap, &st->bg_color);
  pixel = get_pixel_resource (dpy, st->xgwa.colormap,
                              "foreground", "Foreground");
  st->fg_color.pixel = pixel;
  XQueryColor (dpy, st->xgwa.colormap, &st->fg_color);
  set_gc_xcolor (st, st->bg_gc, &st->bg_color);

  face_init[st->face_type] (st);

  st->radius = min (st->width, st->height) / 2.0 * st->size / 100.0;
  if (st->pin) {
    st->center.x = i_round (st->width  / 2.0);
    st->center.y = i_round (st->height / 2.0);
  }
  else {
    st->center.x = i_round (st->radius) +
      i_random (i_round (st->width  - st->radius * 2.0));
    st->center.y = i_round (st->radius) +
      i_random (i_round (st->height - st->radius * 2.0));
  }
  st->orient.x = i_random (2) * 2 - 1;
  st->orient.y = i_random (2) * 2 - 1;

  create_pixmap (st);

  return (st);
}

static void anclock_reshape (Display *dpy, Window window, void *closure,
                             unsigned int w, unsigned int h)
{
  state *st;

  st = (state *) closure;
  XGetWindowAttributes (st->dpy, st->window, &st->xgwa);
  st->width  = st->xgwa.width;
  st->height = st->xgwa.height;
  st->radius = min (st->width, st->height) / 2.0 * st->size / 100.0;

  if (st->pin) {
    st->center.x = i_round (st->width  / 2.0);
    st->center.y = i_round (st->height / 2.0);
  }
  else {
    if (st->center.x - st->radius < 0)
      st->center.x = i_round (st->radius);
    if (st->width < st->center.x + st->radius)
      st->center.x = st->width - i_round (st->radius);
    if (st->center.y - st->radius < 0)
      st->center.y = i_round (st->radius);
    if (st->height < st->center.y + st->radius)
      st->center.y = st->height - i_round (st->radius);
  }

  create_pixmap (st);
}

static Bool anclock_event (Display *dpy, Window window, void *closure,
                           XEvent *event)
{
  return False;
}

static ulong anclock_draw (Display *dpy, Window window, void *closure)
{
  state  *st;
  struct  timeval tv;

  st = (state *) closure;
  draw_clock_pixmap (st);
  XCopyArea (st->dpy, st->pixmap, st->window, st->bg_gc, 0, 0,
             st->width, st->height, 0, 0);
  gettimeofday (&tv, NULL);
  return (st->delay - tv.tv_usec % st->delay);
}

static void anclock_free (Display *dpy, Window window, void *closure)
{
  state *st;

  st = (state *) closure;
  XFreePixmap (st->dpy, st->pixmap);
  XFreeGC (st->dpy, st->bg_gc);
  face_free[st->face_type] (st);
  free (st);
}

XSCREENSAVER_MODULE ("AnClock", anclock)
